@@ -26,7 +26,10 @@ from ravendb.documents.indexes.definitions import (
     AutoIndexDefinition,
     AutoIndexFieldOptions,
 )
-from ravendb.documents.indexes.index_creation import AbstractIndexDefinitionBuilder, AbstractIndexCreationTask
+from ravendb.documents.indexes.abstract_index_creation_tasks import (
+    AbstractIndexDefinitionBuilder,
+    AbstractIndexCreationTask,
+)
 from ravendb.documents.indexes.spatial.configuration import AutoSpatialOptions
 from ravendb.documents.operations.attachments import (
     DeleteAttachmentOperation,
@@ -63,15 +66,15 @@ from ravendb.documents.operations.compare_exchange.operations import (
     GetCompareExchangeValuesOperation,
     DeleteCompareExchangeValueOperation,
 )
-from ravendb.documents.operations.configuration import (
+from ravendb.documents.operations.configuration.operations import (
     GetServerWideClientConfigurationOperation,
     PutServerWideClientConfigurationOperation,
     ClientConfiguration,
     GetClientConfigurationOperation,
     PutClientConfigurationOperation,
-    StudioConfiguration,
-    StudioEnvironment,
 )
+from ravendb.documents.operations.configuration.definitions import StudioConfiguration, StudioEnvironment
+
 from ravendb.documents.operations.connection_strings import ConnectionString
 from ravendb.documents.operations.etl.configuration import EtlConfiguration, RavenEtlConfiguration
 from ravendb.documents.operations.etl.olap import OlapEtlConfiguration
@@ -100,7 +103,7 @@ from ravendb.documents.operations.indexes import (
     GetIndexErrorsOperation,
     IndexingStatus,
 )
-from ravendb.documents.operations.lazy.lazy_operation import LazyOperation
+from ravendb.documents.operations.lazy.definition import LazyOperation
 from ravendb.documents.operations.misc import DeleteByQueryOperation, GetOperationStateOperation, QueryOperationOptions
 from ravendb.documents.operations.patch import (
     PatchOperation,
@@ -117,7 +120,7 @@ from ravendb.documents.operations.replication.definitions import (
     ReplicationNode,
     ExternalReplicationBase,
 )
-from ravendb.documents.operations.revisions.configuration import (
+from ravendb.documents.operations.revisions import (
     RevisionsCollectionConfiguration,
     RevisionsConfiguration,
 )
@@ -177,7 +180,9 @@ from ravendb.documents.session.cluster_transaction_operation import ClusterTrans
 from ravendb.documents.session.document_info import DocumentInfo
 from ravendb.documents.session.document_session import DocumentSession
 from ravendb.documents.session.entity_to_json import EntityToJson
-from ravendb.documents.session.in_memory_document_session_operations import InMemoryDocumentSessionOperations
+from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
+    InMemoryDocumentSessionOperations,
+)
 from ravendb.documents.session.loaders.include import IncludeBuilder, IncludeBuilderBase, QueryIncludeBuilder
 from ravendb.documents.session.loaders.loaders import (
     LoaderWithInclude,
