@@ -8,11 +8,12 @@ from typing import Callable, Union, Optional, TYPE_CHECKING, List, Set, Dict
 
 import requests
 
-from ravendb.documents.operations.counters.operation import (
+from ravendb.documents.operations.counters import (
     CounterOperation,
     DocumentCountersOperation,
     CounterOperationType,
 )
+from ravendb.documents.operations.time_series import TimeSeriesOperation
 from ravendb.documents.session.misc import TransactionMode, ForceRevisionStrategy
 from ravendb.http.raven_command import RavenCommand
 from ravendb.http.server_node import ServerNode
@@ -23,7 +24,9 @@ from ravendb.util.util import RaftIdGenerator
 if TYPE_CHECKING:
     from ravendb.documents.conventions import DocumentConventions
     from ravendb.documents.operations.patch import PatchRequest
-    from ravendb.documents.session.in_memory_document_session_operations import InMemoryDocumentSessionOperations
+    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
+        InMemoryDocumentSessionOperations,
+    )
 
 
 class CommandType(Enum):
@@ -234,13 +237,11 @@ class CommandData:
         name: str = None,
         change_vector: str = None,
         command_type: CommandType = None,
-        on_before_save_changes: Callable = None,
     ):
         self._key = key
         self._name = name
         self._change_vector = change_vector
         self._command_type = command_type
-        self._on_before_save_changes = on_before_save_changes
 
     @property
     def key(self) -> str:
@@ -258,11 +259,8 @@ class CommandData:
     def command_type(self) -> CommandType:
         return self._command_type
 
-    @property
-    def on_before_save_changes(
-        self,
-    ) -> Callable[[InMemoryDocumentSessionOperations], None]:
-        return self._on_before_save_changes
+    def on_before_save_changes(self, session: InMemoryDocumentSessionOperations) -> None:
+        pass
 
     @abstractmethod
     def serialize(self, conventions: DocumentConventions) -> dict:
@@ -270,11 +268,15 @@ class CommandData:
 
 
 class DeleteCommandData(CommandData):
-    def __init__(self, key: str, change_vector: str):
-        super(DeleteCommandData, self).__init__(key=key, command_type=CommandType.DELETE)
+    def __init__(self, key: str, change_vector: str, original_change_vector: str = None):
+        super(DeleteCommandData, self).__init__(key=key, command_type=CommandType.DELETE, change_vector=change_vector)
+        self.original_change_vector = original_change_vector
 
     def serialize(self, conventions: DocumentConventions) -> dict:
-        return {"Id": self.key, "ChangeVector": self.change_vector, "Type": CommandType.DELETE}
+        data = {"Id": self.key, "ChangeVector": self.change_vector, "Type": CommandType.DELETE}
+        if self.original_change_vector is not None:
+            data.update({"OriginalChangeVector": self.original_change_vector})
+        return data
 
 
 class PutCommandDataBase(CommandData):
@@ -492,7 +494,7 @@ class CountersBatchCommandData(CommandData):
 
         super().__init__(key=document_id, command_type=CommandType.COUNTERS, change_vector=change_vector)
         self._from_etl = from_etl
-        self._counters = DocumentCountersOperation(self.key, *counter_operations)
+        self._counters = DocumentCountersOperation(self.key, counter_operations)
 
     def has_increment(self, counter_name):
         self.has_operation_of_type(CounterOperationType.INCREMENT, counter_name)
@@ -517,7 +519,7 @@ class CountersBatchCommandData(CommandData):
         json_dict = {
             "Id": self.key,
             "Counters": self.counters.to_json(),
-            "Type": self.type,
+            "Type": self.command_type,
         }
         if self._from_etl:
             json_dict["FromEtl"] = self._from_etl
@@ -560,7 +562,7 @@ class CopyAttachmentCommandData(CommandData):
         source_name: str,
         destination_document_id: str,
         destination_name: str,
-        change_vector: str,
+        change_vector: Optional[str],
     ):
         if source_document_id.isspace():
             raise ValueError("source_document_id is required")
@@ -592,7 +594,7 @@ class MoveAttachmentCommandData(CommandData):
         name: str,
         destination_id: str,
         destination_name: str,
-        change_vector: str,
+        change_vector: Optional[str],
     ):
         if key.isspace():
             raise ValueError("source_document_id is required")
@@ -618,7 +620,7 @@ class MoveAttachmentCommandData(CommandData):
 
 
 class DeleteAttachmentCommandData(CommandData):
-    def __init__(self, document_id: str, name: str, change_vector: str):
+    def __init__(self, document_id: str, name: str, change_vector: Optional[str]):
         if not document_id:
             raise ValueError(document_id)
         if not name:
@@ -629,6 +631,37 @@ class DeleteAttachmentCommandData(CommandData):
         return {"Id": self._key, "Name": self.name, "ChangeVector": self.change_vector, "Type": self.command_type}
 
 
+class TimeSeriesBatchCommandData(CommandData):
+    def __init__(
+        self,
+        document_id: str,
+        name: str,
+        appends: Optional[List[TimeSeriesOperation.AppendOperation]],
+        deletes: Optional[List[TimeSeriesOperation.DeleteOperation]],
+    ):
+        if not document_id:
+            raise ValueError("Document id cannot be None")
+        if not name:
+            raise ValueError("Name cannot be None")
+
+        super(TimeSeriesBatchCommandData, self).__init__(document_id, name, None, CommandType.TIME_SERIES)
+        self.time_series = TimeSeriesOperation()
+        self.time_series.name = name
+
+        if appends:
+            for append_operation in appends:
+                self.time_series.append(append_operation)
+        if deletes:
+            for delete_operation in deletes:
+                self.time_series.delete(delete_operation)
+
+    def serialize(self, conventions: DocumentConventions) -> dict:
+        return {"Id": self.key, "TimeSeries": self.time_series.to_json(), "Type": "TimeSeries"}
+
+    def on_before_save_changes(self, session: InMemoryDocumentSessionOperations) -> None:
+        pass
+
+
 # ------------ OPTIONS ------------
 
 
