@@ -11,8 +11,12 @@ import inflect
 
 from typing import TypeVar
 from ravendb.json.metadata_as_dictionary import MetadataAsDictionary
-from ravendb import constants
-from ravendb.documents.operations.configuration import ClientConfiguration, LoadBalanceBehavior, ReadBalanceBehavior
+from ravendb.primitives import constants
+from ravendb.documents.operations.configuration.definitions import (
+    ClientConfiguration,
+    LoadBalanceBehavior,
+    ReadBalanceBehavior,
+)
 from ravendb.documents.indexes.definitions import SortOptions
 from ravendb.tools.utils import Utils
 
@@ -23,6 +27,10 @@ _T = TypeVar("_T")
 
 
 class DocumentConventions(object):
+    @classmethod
+    def default_conventions(cls):
+        return cls()
+
     __cached_default_type_collection_names: Dict[type, str] = {}
 
     def __init__(self):
@@ -33,6 +41,7 @@ class DocumentConventions(object):
         self.max_number_of_requests_per_session = 30
         self._max_http_cache_size = 128 * 1024 * 1024
         self.max_length_of_query_using_get_url = 1024 + 512
+        self.time_series_batch_size = 1024
 
         # Flags
         self.disable_topology_updates = False
@@ -40,6 +49,7 @@ class DocumentConventions(object):
         self.throw_if_query_page_size_is_not_set = False
         self._send_application_identifier = True
         self._save_enums_as_integers: Optional[bool] = None
+        self._disable_atomic_document_writes_in_cluster_wide_transaction: Optional[bool] = None
 
         # Configuration
         self.json_default_method = DocumentConventions.json_default
@@ -198,6 +208,15 @@ class DocumentConventions(object):
     def send_application_identifier(self) -> bool:
         return self._send_application_identifier
 
+    @property
+    def disable_atomic_document_writes_in_cluster_wide_transaction(self) -> bool:
+        return self._disable_atomic_document_writes_in_cluster_wide_transaction
+
+    @disable_atomic_document_writes_in_cluster_wide_transaction.setter
+    def disable_atomic_document_writes_in_cluster_wide_transaction(self, value: bool):
+        self.__assert_not_frozen()
+        self._disable_atomic_document_writes_in_cluster_wide_transaction = value
+
     @staticmethod
     def json_default(o):
         if o is None:
