@@ -16,7 +16,7 @@ class LoaderWithInclude:
         pass
 
     @abstractmethod
-    def load(self, object_type: type, *ids: str) -> Dict[str, object]:
+    def load(self, object_type: _T, *ids: str) -> _T:
         pass
 
 
@@ -29,7 +29,7 @@ class MultiLoaderWithInclude(LoaderWithInclude):
         self.__includes.append(path)
         return self
 
-    def load(self, object_type: Type[_T], *ids: str) -> Dict[str, object]:
+    def load(self, object_type: Type[_T], *ids: str) -> Dict[str, _T]:
         return self.__session._load_internal(object_type, list(ids), self.__includes)
 
 
@@ -42,5 +42,5 @@ class LazyMultiLoaderWithInclude(LoaderWithInclude):
         self.__includes.append(path)
         return self
 
-    def load(self, object_type: Type[_T], *ids: str) -> Lazy[Dict[str, object]]:
+    def load(self, object_type: Type[_T], *ids: str) -> Lazy[Dict[str, _T]]:
         return self.__session.lazy_load_internal(object_type, list(ids), self.__includes, None)
