@@ -9,7 +9,7 @@ from ravendb.documents.operations.compare_exchange.compare_exchange_value_result
     CompareExchangeValueResultParser,
 )
 
-from ravendb import constants
+from ravendb.primitives import constants
 from ravendb.documents.queries.facets.misc import FacetResult
 from ravendb.documents.queries.index_query import IndexQuery
 from ravendb.documents.session.document_info import DocumentInfo
@@ -20,15 +20,17 @@ from ravendb.exceptions.raven_exceptions import RavenException
 from ravendb.tools.utils import Utils, CaseInsensitiveDict
 from ravendb.documents.queries.query import QueryResult
 from ravendb.extensions.json_extensions import JsonExtensions
-from ravendb.documents.commands.crud import GetDocumentsResult, ConditionalGetDocumentsCommand, ConditionalGetResult
+from ravendb.documents.commands.crud import GetDocumentsResult, ConditionalGetResult
 from ravendb.documents.session.operations.load_operation import LoadOperation
 from ravendb.documents.conventions import DocumentConventions
-from ravendb.documents.operations.lazy.lazy_operation import LazyOperation
+from ravendb.documents.operations.lazy.definition import LazyOperation
 from ravendb.documents.commands.multi_get import Content, GetRequest, GetResponse
 
 if TYPE_CHECKING:
     from ravendb.documents.session.operations.query import QueryOperation
-    from ravendb.documents.session.in_memory_document_session_operations import InMemoryDocumentSessionOperations
+    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
+        InMemoryDocumentSessionOperations,
+    )
     from ravendb.documents.session.document_session import DocumentSession
     from ravendb.documents.session.cluster_transaction_operation import (
         ClusterTransactionOperationsBase,
@@ -69,11 +71,11 @@ class LazyStartsWithOperation(LazyOperation):
 
         self.result = None
         self.query_result: Union[None, QueryResult] = None
-        self.requires_retry: Union[None, bool] = None
+        self._requires_retry: Union[None, bool] = None
 
     @property
-    def is_requires_retry(self) -> bool:
-        return self.requires_retry
+    def requires_retry(self) -> bool:
+        return self._requires_retry
 
     def create_request(self) -> GetRequest:
         request = GetRequest()
@@ -134,7 +136,7 @@ class LazyConditionalLoadOperation(LazyOperation):
         raise NotImplementedError()
 
     @property
-    def is_requires_retry(self) -> bool:
+    def requires_retry(self) -> bool:
         return self.__requires_retry
 
     def create_request(self) -> GetRequest:
@@ -144,7 +146,7 @@ class LazyConditionalLoadOperation(LazyOperation):
         request.method = "GET"
         request.query = f"?id={Utils.quote_key(self.__key)}"
 
-        request.headers["If-None-Match"] = f'"{self.__change_vector}"'
+        request.headers[constants.Headers.IF_NONE_MATCH] = f'"{self.__change_vector}"'
         return request
 
     def handle_response(self, response: "GetResponse") -> None:
@@ -210,8 +212,8 @@ class LazySessionOperations:
 
     def load_starting_with(
         self,
-        object_type: Type[_T],
         id_prefix: str,
+        object_type: Optional[Type[_T]],
         matches: str = None,
         start: int = 0,
         page_size: int = 25,
@@ -283,7 +285,7 @@ class LazyLoadOperation(LazyOperation):
         self.__result = value
 
     @property
-    def is_requires_retry(self) -> bool:
+    def requires_retry(self) -> bool:
         return self.__requires_retry
 
     def create_request(self) -> Optional[GetRequest]:
@@ -342,7 +344,7 @@ class LazyLoadOperation(LazyOperation):
             LoadOperation(self.__session).by_keys(self.__already_in_session).get_documents(self.__object_type)
 
         self.__load_operation.set_result(load_result)
-        if not self.is_requires_retry:
+        if not self.requires_retry:
             self.result = self.__load_operation.get_documents(self.__object_type)
 
 
@@ -392,7 +394,7 @@ class LazyQueryOperation(Generic[_T], LazyOperation[_T]):
         self.__query_result = value
 
     @property
-    def is_requires_retry(self) -> bool:
+    def requires_retry(self) -> bool:
         return self.__requires_retry
 
     def handle_response(self, response: "GetResponse") -> None:
@@ -458,7 +460,7 @@ class LazyAggregationQueryOperation(LazyOperation):
         self.query_result = query_result
 
     @property
-    def is_requires_retry(self) -> bool:
+    def requires_retry(self) -> bool:
         return self.__requires_retry
 
 
@@ -504,7 +506,7 @@ class LazySuggestionQueryOperation(LazyOperation):
         self.query_result = query_result
 
     @property
-    def is_requires_retry(self) -> bool:
+    def requires_retry(self) -> bool:
         return self.__requires_retry
 
 
@@ -540,7 +542,7 @@ class LazyGetCompareExchangeValueOperation(Generic[_T], LazyOperation[_T]):
         raise NotImplementedError("Not implemented")
 
     @property
-    def is_requires_retry(self) -> bool:
+    def requires_retry(self) -> bool:
         return self.__requires_retry
 
     def create_request(self) -> Optional["GetRequest"]:
@@ -634,7 +636,7 @@ class LazyGetCompareExchangeValuesOperation(LazyOperation[_T], Generic[_T]):
         raise NotImplementedError("Not implemented.")
 
     @property
-    def is_requires_retry(self) -> bool:
+    def requires_retry(self) -> bool:
         return self.__requires_retry
 
     def create_request(self) -> Optional["GetRequest"]:
