@@ -1,9 +1,10 @@
 import datetime
 import enum
+import inspect
 import logging
 from typing import Union, Optional, TypeVar, List, Type, Callable, Dict, TYPE_CHECKING
 
-from ravendb import constants
+from ravendb.primitives import constants
 from ravendb.documents.commands.query import QueryCommand
 from ravendb.documents.queries.index_query import IndexQuery
 from ravendb.documents.queries.query import QueryResult
@@ -13,7 +14,9 @@ from ravendb.exceptions.documents.indexes import IndexDoesNotExistException
 from ravendb.tools.utils import Stopwatch, Utils
 
 if TYPE_CHECKING:
-    from ravendb.documents.session.in_memory_document_session_operations import InMemoryDocumentSessionOperations
+    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
+        InMemoryDocumentSessionOperations,
+    )
     from ravendb.documents.session.document_session import DocumentSession
 
 
@@ -130,12 +133,11 @@ class QueryOperation:
                 query_result.results, query_result.includes, query_result.included_paths
             )
 
-            # todo: counters and timeseries
             if query_result.counter_includes is not None:
-                raise NotImplementedError()
+                self.__session.register_counters(query_result.counter_includes, query_result.included_counter_names)
 
             if query_result.time_series_includes is not None:
-                raise NotImplementedError()
+                self.__session.register_time_series(query_result.time_series_includes)
 
             if query_result.compare_exchange_value_includes is not None:
                 self.__session.cluster_transaction.register_compare_exchange_values(
@@ -198,7 +200,12 @@ class QueryOperation:
             BeforeConversionToEntityEventArgs(session, key, object_type, document)
         )
 
-        result = Utils.initialize_object(document, object_type)
+        # By custom defined 'from_json' serializer class method
+        # todo: make separate interface to do from_json
+        if "from_json" in object_type.__dict__ and inspect.ismethod(object_type.from_json):
+            result = object_type.from_json(document)
+        else:
+            result = Utils.initialize_object(document, object_type)
         session.after_conversion_to_entity_invoke(AfterConversionToEntityEventArgs(session, key, document, result))
 
         return result
