@@ -19,7 +19,9 @@ from typing import (
     TYPE_CHECKING,
 )
 
-from ravendb import constants
+from ravendb.documents.queries.time_series import TimeSeriesQueryBuilder
+from ravendb.documents.session.time_series import TimeSeriesRange, ITimeSeriesValuesBindable
+from ravendb.primitives import constants
 from ravendb.documents.conventions import DocumentConventions
 from ravendb.documents.indexes.spatial.configuration import SpatialUnits, SpatialRelation
 from ravendb.documents.queries.explanation import Explanations, ExplanationOptions
@@ -82,6 +84,8 @@ from ravendb.documents.session.tokens.query_tokens.definitions import (
     DistinctToken,
     ShapeToken,
     SuggestToken,
+    CounterIncludesToken,
+    TimeSeriesIncludesToken,
 )
 from ravendb.documents.session.utils.document_query import DocumentQueryHelper
 from ravendb.documents.session.utils.includes_util import IncludesUtil
@@ -90,6 +94,7 @@ from ravendb.tools.utils import Utils
 _T = TypeVar("_T")
 _TResult = TypeVar("_TResult")
 _TProjection = TypeVar("_TProjection")
+_T_TS_Bindable = TypeVar("_T_TS_Bindable", bound=ITimeSeriesValuesBindable)
 if TYPE_CHECKING:
     from ravendb.documents.store.definition import Lazy
     from ravendb.documents.session.document_session import DocumentSession
@@ -101,11 +106,11 @@ class AbstractDocumentQuery(Generic[_T]):
         self,
         object_type: Type[_T],
         session: "InMemoryDocumentSessionOperations",
-        index_name: Union[None, str],
-        collection_name: Union[None, str],
+        index_name: Optional[str],
+        collection_name: Optional[str],
         is_group_by: bool,
-        declare_tokens: Union[None, List[DeclareToken]],
-        load_tokens: Union[None, List[LoadToken]],
+        declare_tokens: Optional[List[DeclareToken]],
+        load_tokens: Optional[List[LoadToken]],
         from_alias: Optional[str] = None,
         is_project_into: Optional[bool] = False,
     ):
@@ -121,21 +126,21 @@ class AbstractDocumentQuery(Generic[_T]):
         self.__conventions = DocumentConventions() if session is None else session.conventions
         self.is_project_into = is_project_into if is_project_into is not None else False
 
-        self._is_intersect: Union[None, bool] = None
-        self.__is_in_more_like_this: Union[None, bool] = None
-        self._includes_alias: Union[None, str] = None
-        self._negate: Union[None, bool] = None
-        self._query_raw: Union[None, str] = None
+        self._is_intersect: Optional[bool] = None
+        self.__is_in_more_like_this: Optional[bool] = None
+        self._includes_alias: Optional[str] = None
+        self._negate: Optional[bool] = None
+        self._query_raw: Optional[str] = None
         self._query_parameters: Dict[str, object] = {}
         self.__alias_to_group_by_field_name: Dict[str, str] = {}
 
         self._document_includes: Set[str] = set()
-        self._page_size: Union[None, int] = None
-        self._start: Union[None, int] = None
+        self._page_size: Optional[int] = None
+        self._start: Optional[int] = None
         self.__current_clause_depth: int = 0
 
         self._select_tokens: List[QueryToken] = []
-        self._fields_to_fetch_token: Union[None, FieldsToFetchToken] = None
+        self._fields_to_fetch_token: Optional[FieldsToFetchToken] = None
         self._where_tokens: List[QueryToken] = []
         self._group_by_tokens: List[QueryToken] = []
         self._order_by_tokens: List[QueryToken] = []
@@ -143,8 +148,10 @@ class AbstractDocumentQuery(Generic[_T]):
         self._compare_exchange_includes_tokens: List[CompareExchangeValueIncludesToken] = []
         self._highlighting_tokens: List[HighlightingToken] = []
         self._query_highlightings: QueryHighlightings = QueryHighlightings()
-        self._explanation_token: Union[None, ExplanationToken] = None
-        self._explanations: Union[None, Explanations] = None
+        self._explanation_token: Optional[ExplanationToken] = None
+        self._explanations: Optional[Explanations] = None
+        self._counter_includes_tokens: Optional[List[CounterIncludesToken]] = None
+        self._time_series_includes_tokens: Optional[List[TimeSeriesIncludesToken]] = None
 
         self._before_query_executed_callback: List[Callable[[IndexQuery], None]] = []
         self._after_query_executed_callback: List[Callable[[QueryResult], None]] = [
@@ -153,16 +160,16 @@ class AbstractDocumentQuery(Generic[_T]):
         self._after_stream_executed_callback: List[Callable[[Dict], None]] = []
 
         self._query_stats = QueryStatistics()
-        self._disable_entities_tracking: Union[None, bool] = None
-        self._disable_caching: Union[None, bool] = None
-        self._projection_behavior: Union[None, ProjectionBehavior] = None
+        self._disable_entities_tracking: Optional[bool] = None
+        self._disable_caching: Optional[bool] = None
+        self._projection_behavior: Optional[ProjectionBehavior] = None
         self.parameter_prefix = "p"
-        self._query_timings: Union[None, QueryTimings] = None
+        self._query_timings: Optional[QueryTimings] = None
         self._default_operator = QueryOperator.AND
-        self._the_wait_for_non_stale_results: Union[None, bool] = None
-        self._timeout: Union[None, datetime.timedelta] = None
+        self._the_wait_for_non_stale_results: Optional[bool] = None
+        self._timeout: Optional[datetime.timedelta] = None
 
-        self._query_operation: Union[None, QueryOperation] = None
+        self._query_operation: Optional[QueryOperation] = None
 
     @property
     def conventions(self) -> DocumentConventions:
@@ -211,7 +218,7 @@ class AbstractDocumentQuery(Generic[_T]):
 
         self._default_operator = operator
 
-    def _wait_for_non_stale_results(self, wait_timeout: Union[None, datetime.timedelta]) -> None:
+    def _wait_for_non_stale_results(self, wait_timeout: Optional[datetime.timedelta]) -> None:
         # Graph queries may set this property multiple times
         if self._the_wait_for_non_stale_results:
             if self._timeout is None or (wait_timeout is not None and self._timeout.seconds < wait_timeout.seconds):
@@ -363,15 +370,18 @@ class AbstractDocumentQuery(Generic[_T]):
             includes = path_or_include_builder
             if includes is None:
                 return
-            if includes._documents_to_include is not None:
-                self._document_includes.update(includes._documents_to_include)
+            if includes.documents_to_include is not None:
+                self._document_includes.update(includes.documents_to_include)
 
-            # todo: counters and time series includes
-            # self._include_counters(includes.alias, includes.counters_to_include_by_source_path)
-            # if includes.time_series_to_include_by_source_alias is not None:
-            #     self._include_time_series(includes.alias, includes.time_series_to_include_by_source_alias)
-            if includes._compare_exchange_values_to_include is not None:
-                for compare_exchange_value in includes._compare_exchange_values_to_include:
+            self._include_counters(includes.alias, includes.counters_to_include_by_source_path)
+
+            if includes.time_series_to_include_by_source_alias is not None:
+                self._include_time_series(includes.alias, includes.time_series_to_include_by_source_alias)
+
+            if includes.compare_exchange_values_to_include is not None:
+                self._compare_exchange_includes_tokens = []
+
+                for compare_exchange_value in includes.compare_exchange_values_to_include:
                     self._compare_exchange_includes_tokens.append(
                         CompareExchangeValueIncludesToken.create(compare_exchange_value)
                     )
@@ -461,7 +471,7 @@ class AbstractDocumentQuery(Generic[_T]):
             for arg in mc.args:
                 args.append(self.__add_query_parameter(arg))
 
-            token: Union[None, WhereToken] = None
+            token: Optional[WhereToken] = None
             object_type = type(mc)
             if object_type == CmpXchg:
                 token = WhereToken.create(
@@ -918,8 +928,8 @@ class AbstractDocumentQuery(Generic[_T]):
             and self._explanation_token is None
             and self._query_timings is None
             and not self._compare_exchange_includes_tokens
-            # todo: and self._counter_includes_tokens is None
-            # todo: and self._time_series_includes_tokens is None
+            and self._counter_includes_tokens is None
+            and self._time_series_includes_tokens is None
         ):
             return
 
@@ -938,9 +948,8 @@ class AbstractDocumentQuery(Generic[_T]):
             else:
                 query_text.append(include)
 
-        # todo: counters & time series
-        # first = self.__write_include_tokens(self._counter_includes_tokens, first, query_text)
-        # first = self.__write_include_tokens(self._time_series_includes_tokens, first, query_text)
+        first = self.__write_include_tokens(self._counter_includes_tokens, first, query_text)
+        first = self.__write_include_tokens(self._time_series_includes_tokens, first, query_text)
         first = self.__write_include_tokens(self._compare_exchange_includes_tokens, first, query_text)
         first = self.__write_include_tokens(self._highlighting_tokens, first, query_text)
 
@@ -1167,7 +1176,7 @@ class AbstractDocumentQuery(Generic[_T]):
                 result.append(self.__transform_value(nested_where_params))
         return result
 
-    def __negate_if_needed(self, tokens: List[QueryToken], field_name: Union[None, str]) -> None:
+    def __negate_if_needed(self, tokens: List[QueryToken], field_name: Optional[str]) -> None:
         if not self._negate:
             return
 
@@ -1260,12 +1269,18 @@ class AbstractDocumentQuery(Generic[_T]):
             from_alias = self._includes_alias
             self.add_from_alias_to_where_tokens(from_alias)
 
-        # todo: counter and time series tokens
+        if self._counter_includes_tokens is not None:
+            for counter_includes_token in self._counter_includes_tokens:
+                counter_includes_token.add_alias_to_path(from_alias)
+
+        if self._time_series_includes_tokens is not None:
+            for time_series_include_token in self._time_series_includes_tokens:
+                time_series_include_token.add_alias_to_path(from_alias)
 
         return from_alias
 
     @staticmethod
-    def _get_source_alias_if_exists(object_type: type, query_data: QueryData, fields: List[str]) -> Union[None, str]:
+    def _get_source_alias_if_exists(object_type: type, query_data: QueryData, fields: List[str]) -> Optional[str]:
         if len(fields) != 1 or fields[0] is None:
             return None
         try:
@@ -1285,8 +1300,13 @@ class AbstractDocumentQuery(Generic[_T]):
 
         return possible_alias
 
-    # todo: def _create_time_series_query_data(time_series_query: Callable[[TimeSeriesQueryBuilder], None])
-    #  -> QueryData:
+    def _create_time_series_query_data(self, time_series_query: Callable[[TimeSeriesQueryBuilder], None]) -> QueryData:
+        builder = TimeSeriesQueryBuilder()
+        time_series_query(builder)
+
+        fields = [constants.TimeSeries.SELECT_FIELD_NAME + "(" + builder.query_text + ")"]
+        projections = [constants.TimeSeries.QUERY_FUNCTION]
+        return QueryData(fields, projections)
 
     def _add_before_query_executed_listener(self, action: Callable[[IndexQuery], None]) -> None:
         self._before_query_executed_callback.append(action)
@@ -1556,7 +1576,7 @@ class AbstractDocumentQuery(Generic[_T]):
     def __iter__(self) -> Iterator[_T]:
         return self.__execute_query_operation(None).__iter__()
 
-    def __execute_query_operation(self, take: Union[None, int]) -> List[_T]:
+    def __execute_query_operation(self, take: Optional[int]) -> List[_T]:
         self.__execute_query_operation_internal(take)
 
         return self._query_operation.complete(self._object_type)
@@ -1572,6 +1592,12 @@ class AbstractDocumentQuery(Generic[_T]):
 
         return self._query_operation.current_query_results.create_snapshot()
 
+    def single(self) -> _T:
+        result = list(self.__execute_query_operation(2))
+        if len(result) != 1:
+            raise ValueError(f"Expected single result, got: {len(result)} ")
+        return result[0]
+
     def _aggregate_by(self, facet: FacetBase) -> None:
         for token in self._select_tokens:
             if isinstance(token, FacetToken):
@@ -1603,7 +1629,7 @@ class AbstractDocumentQuery(Generic[_T]):
         self._the_session: "DocumentSession"
         return self._the_session.add_lazy_count_operation(self.lazy_query_operation)
 
-    def _suggest_using(self, suggestion: Union[None, SuggestionBase]) -> None:
+    def _suggest_using(self, suggestion: Optional[SuggestionBase]) -> None:
         if suggestion is None:
             raise ValueError("suggestion cannot be None")
 
@@ -1664,13 +1690,37 @@ class AbstractDocumentQuery(Generic[_T]):
         self._explanations = Explanations()
         explanations_callback(self._explanations)
 
-    # todo: counters
-    def _include_counters(self, alias: str, counter_to_include_by_doc_id: Dict[str, Tuple[bool, Set[str]]]) -> None:
-        raise NotImplementedError()
+    def _include_counters(
+        self, alias: str, counter_to_include_by_doc_id: Optional[Dict[str, Tuple[bool, Set[str]]]]
+    ) -> None:
+        if not counter_to_include_by_doc_id:
+            return
 
-    # todo: time series
-    # def _include_time_series(self, alias:str, time_series_to_include:Dict[str,Set[AbstractTimeSeriesRange]]) -> None:
-    # pass
+        self._counter_includes_tokens = []
+        self._includes_alias = alias
+
+        for key, value in counter_to_include_by_doc_id.items():
+            if value[0]:
+                self._counter_includes_tokens.append(CounterIncludesToken.all(key))
+                continue
+
+            if not value[1]:
+                continue
+
+            for name in value[1]:
+                self._counter_includes_tokens.append(CounterIncludesToken.create(key, name))
+
+    def _include_time_series(self, alias: str, time_series_to_include: Dict[str, Set[TimeSeriesRange]]) -> None:
+        if not time_series_to_include:
+            return
+
+        self._time_series_includes_tokens = []
+        if self._includes_alias is None:
+            self._includes_alias = alias
+
+        for ts_name, ranges in time_series_to_include.items():
+            for ts_range in ranges:
+                self._time_series_includes_tokens.append(TimeSeriesIncludesToken.create(ts_name, ts_range))
 
 
 class DocumentQuery(Generic[_T], AbstractDocumentQuery[_T]):
@@ -1698,7 +1748,11 @@ class DocumentQuery(Generic[_T], AbstractDocumentQuery[_T]):
             is_project_into,
         )
 
-    # todo: selectTimeSeries
+    def select_time_series(
+        self, projection_class: Type[_T_TS_Bindable], time_series_query: Callable[[TimeSeriesQueryBuilder], None]
+    ) -> DocumentQuery[_T_TS_Bindable]:
+        query_data = self._create_time_series_query_data(time_series_query)
+        return self.select_fields_query_data(projection_class, query_data)
 
     def distinct(self) -> DocumentQuery[_T]:
         self._distinct()
@@ -1741,8 +1795,7 @@ class DocumentQuery(Generic[_T], AbstractDocumentQuery[_T]):
         query_data.project_into = True
         query_data.projection_behavior = projection_behavior
 
-        select_fields = self.select_fields_query_data(projection_class, query_data=query_data)
-        return select_fields
+        return self.select_fields_query_data(projection_class, query_data=query_data)
 
     def wait_for_non_stale_results(self, wait_timeout: datetime.timedelta = None) -> DocumentQuery[_T]:
         self._wait_for_non_stale_results(wait_timeout)
@@ -1836,7 +1889,7 @@ class DocumentQuery(Generic[_T], AbstractDocumentQuery[_T]):
         self.negate_next()
         return self
 
-    def take(self, count: int) -> DocumentQuery[_T]:
+    def take(self: DocumentQuery[_T], count: int) -> DocumentQuery[_T]:
         self._take(count)
         return self
 
@@ -1852,12 +1905,6 @@ class DocumentQuery(Generic[_T], AbstractDocumentQuery[_T]):
         query_result = self.get_query_result()
         return query_result.total_results
 
-    def single(self) -> _T:
-        result = list(self.take(2))
-        if len(result) != 1:
-            raise ValueError(f"Expected signle result, got: {len(result)} ")
-        return result[0]
-
     def where_lucene(self, field_name: str, where_clause: str, exact: bool = False) -> DocumentQuery[_T]:
         self._where_lucene(field_name, where_clause, exact)
         return self
@@ -2041,7 +2088,7 @@ class DocumentQuery(Generic[_T], AbstractDocumentQuery[_T]):
         query._the_wait_for_non_stale_results = self._the_wait_for_non_stale_results
         query._negate = self._negate
         query._document_includes = self._document_includes
-        # todo: self._counter_includes_tokens = self._counter_includes_tokens
+        query._counter_includes_tokens = self._counter_includes_tokens
         # todo: self._time_series_includes_tokens = self._time_series_includes_tokens
         query._compare_exchange_includes_tokens = self._compare_exchange_includes_tokens
         query._root_types = {self._object_type}
@@ -2077,7 +2124,7 @@ class DocumentQuery(Generic[_T], AbstractDocumentQuery[_T]):
 
         return AggregationDocumentQuery(self)
 
-    def aggregate_by_facets(self, *facets: FacetBase) -> AggregationDocumentQuery[_T]:
+    def aggregate_by_facets(self, facets: List[FacetBase]) -> AggregationDocumentQuery[_T]:
         for facet in facets:
             self.aggregate_by(facet)
 
@@ -2213,6 +2260,9 @@ class RawDocumentQuery(Generic[_T], AbstractDocumentQuery[_T]):
         self._take(count)
         return self
 
+    def first(self) -> _T:
+        return list(self.take(1))[0]
+
     def wait_for_non_stale_results(self, wait_timeout: Optional[datetime.timedelta] = None):
         self._wait_for_non_stale_results(wait_timeout)
         return self
@@ -2346,9 +2396,9 @@ class WhereParams:
         self.nested_path = False
         self.allow_wildcards = False
 
-        self.field_name: Union[None, str] = None
-        self.value: Union[None, object] = None
-        self.exact: Union[None, bool] = None
+        self.field_name: Optional[str] = None
+        self.value: Optional[object] = None
+        self.exact: Optional[bool] = None
 
 
 class QueryStatistics:
