@@ -4,9 +4,15 @@ import enum
 import os
 from typing import List, Union, Optional, Tuple
 
-from ravendb import constants
+from ravendb.documents.session.time_series import (
+    TimeSeriesRange,
+    TimeSeriesTimeRange,
+    TimeSeriesRangeType,
+    AbstractTimeSeriesRange,
+    TimeSeriesCountRange,
+)
+from ravendb.primitives import constants
 from ravendb.documents.session.misc import OrderingType
-from ravendb.data.timeseries import TimeSeriesRange
 from ravendb.documents.indexes.spatial.configuration import SpatialUnits
 from ravendb.documents.queries.group_by import GroupByMethod
 from ravendb.documents.queries.misc import SearchOperator
@@ -63,12 +69,12 @@ class CounterIncludesToken(QueryToken):
 
 
 class TimeSeriesIncludesToken(QueryToken):
-    def __init__(self, source_path: str, time_range: TimeSeriesRange):
+    def __init__(self, source_path: str, time_range: AbstractTimeSeriesRange):
         self.__range = time_range
         self.__source_path = source_path
 
     @classmethod
-    def create(cls, source_path: str, time_range: TimeSeriesRange) -> TimeSeriesIncludesToken:
+    def create(cls, source_path: str, time_range: AbstractTimeSeriesRange) -> TimeSeriesIncludesToken:
         return cls(source_path, time_range)
 
     def add_alias_to_path(self, alias: str):
@@ -79,9 +85,43 @@ class TimeSeriesIncludesToken(QueryToken):
         if self.__source_path:
             writer.append(f"{self.__source_path}, ")
 
-        writer.append(f"'{self.__range.name}', ")
-        writer.append(f"'{Utils.datetime_to_string(self.__range.from_date)}', " if self.__range.from_date else "null,")
-        writer.append(f"'{Utils.datetime_to_string(self.__range.to_date)}', " if self.__range.to_date else "null")
+        if self.__range.name:
+            writer.append(f"'{self.__range.name}', ")
+
+        if isinstance(self.__range, TimeSeriesRange):
+            self.__range: TimeSeriesRange
+
+            writer.append(
+                f"'{Utils.datetime_to_string(self.__range.from_date)}', " if self.__range.from_date else "null,"
+            )
+            writer.append(f"'{Utils.datetime_to_string(self.__range.to_date)}', " if self.__range.to_date else "null")
+
+        elif isinstance(self.__range, TimeSeriesTimeRange):
+            self.__range: TimeSeriesTimeRange
+
+            if self.__range.type == TimeSeriesRangeType.LAST:
+                writer.append("last(")
+            else:
+                raise ValueError(f"Not supported time range type: {str(self.__range.type)}")
+            writer.append(str(self.__range.time.value))
+            writer.append(", '")
+            writer.append(self.__range.time.unit.value)
+            writer.append("')")
+
+        elif isinstance(self.__range, TimeSeriesCountRange):
+            self.__range: TimeSeriesCountRange
+
+            if self.__range.type == TimeSeriesRangeType.LAST:
+                writer.append("last(")
+            else:
+                raise ValueError(f"Not supported time range type: {str(self.__range.type)}")
+
+            writer.append(str(self.__range.count))
+            writer.append(")")
+
+        else:
+            raise ValueError(f"Not supported time range type: {str(self.__range.__class__.__name__)}")
+
         writer.append(")")
 
 
@@ -234,7 +274,10 @@ class OrderByToken(QueryToken):
     def __init__(self, field_name: str, descending: bool, ordering_or_sorter_name: Union[OrderingType, str]):
         self.__field_name = field_name
         self.__descending = descending
+
+        # duck typing the arg
         is_ordering = isinstance(ordering_or_sorter_name, OrderingType)
+
         self.__ordering = ordering_or_sorter_name if is_ordering else None
         self.__sorter_name = None if is_ordering else ordering_or_sorter_name
 
@@ -362,8 +405,9 @@ class GroupByKeyToken(QueryToken):
         if self.__projected_name is None or self.__projected_name == self.__field_name:
             return
 
-        writer.append(" as ")
+        writer.append(" as '")
         writer.append(self.__projected_name)
+        writer.append("'")
 
 
 class GroupBySumToken(QueryToken):
