@@ -6,9 +6,9 @@ from abc import abstractmethod
 from concurrent.futures import ThreadPoolExecutor
 from typing import Callable, Union, Optional, TypeVar, List, Dict, TYPE_CHECKING
 
-from ravendb import constants, exceptions
 from ravendb.changes.database_changes import DatabaseChanges
-from ravendb.documents.bulk_insert_operation import BulkInsertOperation
+from ravendb.documents.bulk_insert_operation import BulkInsertOperation, BulkInsertOptions
+from ravendb.documents.indexes.index_creation import IndexCreation
 from ravendb.documents.operations.executor import MaintenanceOperationExecutor, OperationExecutor
 from ravendb.documents.operations.indexes import PutIndexesOperation
 from ravendb.documents.session.event_args import (
@@ -28,9 +28,12 @@ from ravendb.documents.session.event_args import (
 )
 from ravendb.documents.store.lazy import Lazy
 from ravendb.documents.session.document_session import DocumentSession
-from ravendb.documents.session.in_memory_document_session_operations import InMemoryDocumentSessionOperations
+from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
+    InMemoryDocumentSessionOperations,
+)
 from ravendb.documents.session.misc import SessionOptions
 from ravendb.documents.subscriptions.document_subscriptions import DocumentSubscriptions
+from ravendb.documents.time_series import TimeSeriesOperations
 from ravendb.http.request_executor import RequestExecutor
 from ravendb.documents.identity.hilo import MultiDatabaseHiLoGenerator
 from ravendb.http.topology import Topology
@@ -40,7 +43,7 @@ from ravendb.documents.conventions import DocumentConventions
 T = TypeVar("T")
 
 if TYPE_CHECKING:
-    from ravendb.documents.indexes.index_creation import AbstractIndexCreationTask, IndexCreation
+    from ravendb.documents.indexes.abstract_index_creation_tasks import AbstractIndexCreationTask
 
 
 class DocumentStoreBase:
@@ -48,12 +51,12 @@ class DocumentStoreBase:
         self.__conventions = None
         self._initialized: bool = False
 
-        self.__certificate_pem_path: Union[None, str] = None
-        self.__trust_store_path: Union[None, str] = None
+        self.__certificate_pem_path: Optional[str] = None
+        self.__trust_store_path: Optional[str] = None
 
         self._urls: List[str] = []
-        self._database: Union[None, str] = None
-        self._disposed: Union[None, bool] = None
+        self._database: Optional[str] = None
+        self._disposed: Optional[bool] = None
 
         self.__before_store: List[Callable[[BeforeStoreEventArgs], None]] = []
         self.__after_save_changes: List[Callable[[AfterSaveChangesEventArgs], None]] = []
@@ -311,14 +314,16 @@ class DocumentStore(DocumentStoreBase):
         self.database = database
         self.__request_executors: Dict[str, Lazy[RequestExecutor]] = CaseInsensitiveDict()
         # todo: aggressive cache
-        self.__maintenance_operation_executor: Union[None, MaintenanceOperationExecutor] = None
-        self.__operation_executor: Union[None, OperationExecutor] = None
+        self.__maintenance_operation_executor: Optional[MaintenanceOperationExecutor] = None
+        self.__operation_executor: Optional[OperationExecutor] = None
         # todo: database smuggler
-        self.__identifier: Union[None, str] = None
+        self.__multi_db_hilo: Optional[MultiDatabaseHiLoGenerator] = None
+        self.__identifier: Optional[str] = None
         self.__add_change_lock = threading.Lock()
-        self.__database_changes = {}
+        self.__database_changes: Dict[str, DatabaseChanges] = {}
         self.__after_close: List[Callable[[], None]] = []
         self.__before_close: List[Callable[[], None]] = []
+        self.__time_series_operation: Optional[TimeSeriesOperations] = None
 
     def __enter__(self):
         return self
@@ -327,7 +332,11 @@ class DocumentStore(DocumentStoreBase):
         self.close()
 
     @property
-    def thread_pool_executor(self):
+    def hilo_id_generator(self) -> Optional[MultiDatabaseHiLoGenerator]:
+        return self.__multi_db_hilo
+
+    @property
+    def thread_pool_executor(self) -> ThreadPoolExecutor:
         return self.__thread_pool_executor
 
     @property
@@ -335,7 +344,7 @@ class DocumentStore(DocumentStoreBase):
         return self.__subscriptions
 
     @property
-    def identifier(self) -> Union[None, str]:
+    def identifier(self) -> Optional[str]:
         if self.__identifier is not None:
             return self.__identifier
 
@@ -364,6 +373,8 @@ class DocumentStore(DocumentStoreBase):
         self.__before_close.remove(event)
 
     def close(self):
+        if self._disposed:
+            return
         for event in self.__before_close:
             event()
 
@@ -377,8 +388,10 @@ class DocumentStore(DocumentStoreBase):
                 self.__multi_db_hilo.return_unused_range()
             except Exception:
                 pass  # ignore
-        # todo: clear subscriptions
-        self._disposed = True
+
+        if self.subscriptions is not None:
+            self.subscriptions.close()
+
         for event in self.__after_close:
             event()
 
@@ -389,16 +402,18 @@ class DocumentStore(DocumentStoreBase):
             lazy.value.close()
 
         self.__thread_pool_executor.shutdown()
+        self._disposed = True
 
     def open_session(
         self, database: Optional[str] = None, session_options: Optional[SessionOptions] = None
     ) -> DocumentSession:
-        if not database and not session_options:
+        if not session_options:
             session_options = SessionOptions()
-        if not ((session_options is not None) ^ (database is not None)):
-            raise ValueError("Pass either database str or session_options object")
-        if database:
-            session_options = SessionOptions(database=database)
+            session_options.database = database
+            session_options.disable_atomic_document_writes_in_cluster_wide_transaction = (
+                self.conventions.disable_atomic_document_writes_in_cluster_wide_transaction
+            )
+
         self.assert_initialized()
         self._ensure_not_closed()
 
@@ -461,7 +476,7 @@ class DocumentStore(DocumentStoreBase):
         self.assert_initialized()
         indexes_to_add = IndexCreation.create_indexes_to_add(tasks, self.conventions)
 
-        self.maintenance.for_database(self.get_effective_database(database)).send(PutIndexesOperation(indexes_to_add))
+        self.maintenance.for_database(self.get_effective_database(database)).send(PutIndexesOperation(*indexes_to_add))
 
     def changes(self, database=None, on_error=None, executor=None) -> DatabaseChanges:  # todo: sync with java
         self.assert_initialized()
@@ -478,8 +493,8 @@ class DocumentStore(DocumentStoreBase):
                 )
             return self.__database_changes[database]
 
-    def __on_close_change(self, database):
-        self.__database_changes.pop(database, None)
+    def __on_close_change(self, database_name: str):
+        self.__database_changes.pop(database_name, None)
 
     def set_request_timeout(self, timeout: datetime.timedelta, database: Optional[str] = None) -> Callable[[], None]:
         self.assert_initialized()
@@ -515,9 +530,9 @@ class DocumentStore(DocumentStoreBase):
 
     # todo: aggressively cache
 
-    def bulk_insert(self, database_name: Optional[str] = None) -> BulkInsertOperation:
+    def bulk_insert(self, database_name: str = None, options: BulkInsertOptions = None) -> BulkInsertOperation:
         self.assert_initialized()
-        return BulkInsertOperation(self.get_effective_database(database_name), self)
+        return BulkInsertOperation(self.get_effective_database(database_name), self, options)
 
     def _assert_valid_configuration(self) -> None:
         if not self.urls:
@@ -538,3 +553,10 @@ class DocumentStore(DocumentStoreBase):
             self.__operation_executor = OperationExecutor(self)
 
         return self.__operation_executor
+
+    @property
+    def time_series(self) -> TimeSeriesOperations:
+        if self.__time_series_operation is None:
+            self.__time_series_operation = TimeSeriesOperations(self)
+
+        return self.__time_series_operation
